#
#  Copyright (c) 2022 IBM Corp.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
import functools
import json
import logging
import re
from typing import List, Mapping, Sequence, Union
import os
from flask import current_app, request, jsonify

from label_sleuth.analysis_utils.analyze_tokens import ngrams_by_info_gain
from label_sleuth.data_access.core.data_structs import TextElement, LabeledTextElement
from label_sleuth.data_access.data_access_api import get_document_uri
from label_sleuth.models.core.languages import Languages
from label_sleuth.orchestrator.core.state_api.orchestrator_state_api import Iteration, IterationStatus
from label_sleuth.orchestrator.orchestrator_api import TRAIN_COUNTS_STR_KEY
from label_sleuth.utils import make_error


def validate_workspace_id(function):
    @functools.wraps(function)
    def wrapper(workspace_id, *args, **kwargs):
        if not current_app.orchestrator_api.workspace_exists(workspace_id):
            return jsonify({"type": "workspace_id_does_not_exist",
                            "title": f"workspace_id {workspace_id} does not exist"}), 404
        return function(workspace_id, *args, **kwargs)

    return wrapper


def validate_category_id(function):
    @functools.wraps(function)
    def wrapper(workspace_id, *args, **kwargs):
        category_id = request.args.get('category_id')
        if category_id is None:
            return jsonify({"type": "missing_category_id", "title": "category_id was not provided"}), 422

        category_id = int(category_id)
        if category_id not in current_app.orchestrator_api.get_all_categories(workspace_id):
            return jsonify({"type": "category_id_does_not_exist",
                            "title": f"category_id {category_id} does not exist in workspace {workspace_id}"}), 404
        return function(workspace_id, *args, **kwargs)

    return wrapper


def elements_back_to_front(workspace_id: str, elements: List[Union[TextElement, LabeledTextElement]], category_id: int,
                           need_snippet: bool = True,
                           query_string: str = None, is_regex: bool = False) -> List[Mapping]:
    """
    Converts TextElement objects from the backend into dictionaries in the form expected by the frontend, and adds
    the model prediction for the elements if available.
    :param workspace_id:
    :param elements: a list of TextElements
    :param category_id:
    :param need_snippet: whether to create a text snippet
    :param query_string: query to include in snippet
    :return: a list of dictionaries with element information
    """

    element_uri_to_info = \
        {text_element.uri:
             {'id': text_element.uri,
              'docid': get_document_uri(text_element.uri),
              'begin': text_element.span[0][0],
              'end': text_element.span[0][1],
              'text': text_element.text,
              'user_labels': {} if type(text_element)==TextElement else {k: str(v.label).lower()
                              # TODO current UI is using true and false as strings. change to boolean in the new UI
                              for k, v in text_element.category_to_label.items()},
              'model_predictions': {}
              }
         for text_element in elements}
    if need_snippet:
        for text_element in elements:
            snippet = get_text_snippet(text_element.text, query_string, is_regex)
            if snippet != element_uri_to_info[text_element.uri]['text']:
                element_uri_to_info[text_element.uri]['snippet'] = snippet

    if category_id is not None and len(elements) > 0 \
            and len(current_app.orchestrator_api.get_all_iterations_by_status(workspace_id, category_id,
                                                                              IterationStatus.READY)) > 0:
        predicted_labels = [pred.label
                            for pred in current_app.orchestrator_api.infer(workspace_id, category_id, elements)]
        for text_element, prediction in zip(elements, predicted_labels):
            # the frontend expects string labels and not boolean
            element_uri_to_info[text_element.uri]['model_predictions'][category_id] = str(prediction).lower()

    return [element_info for element_info in element_uri_to_info.values()]


def get_element(workspace_id, category_id, element_id):
    """
    Get element by id
    :param workspace_id:
    :param category_id:
    :param element_id:
    """
    dataset_name = current_app.orchestrator_api.get_dataset_name(workspace_id)
    element = current_app.orchestrator_api.get_text_elements_by_uris(workspace_id, dataset_name, [element_id])
    element_transformed = elements_back_to_front(workspace_id, element, category_id)[0]
    return element_transformed


def _extract_model_info(iteration):
    # TODO reorganize model information
    placeholder = 'Unknown'
    model_info_dict = {'model_status': placeholder, 'creation_epoch': placeholder, 'model_type': placeholder,
                       'model_metadata': placeholder}
    if iteration.model is not None:
        model_info_dict['model_status'] = iteration.model.model_status.name
        model_info_dict['creation_epoch'] = iteration.model.creation_date.timestamp()
        model_info_dict['model_type'] = iteration.model.model_type.name \
            if iteration.model.model_type is not None else placeholder
        # The frontend expects a dict of string to int, and train counts contains a mix of boolean and string keys.
        model_info_dict['model_metadata'] = \
            {**iteration.model.train_statistics,
             TRAIN_COUNTS_STR_KEY: {str(k).lower(): v for k, v
                                    in iteration.model.train_statistics.get(TRAIN_COUNTS_STR_KEY, {}).items()}}
    return model_info_dict


def extract_iteration_information_list(iterations: Sequence[Iteration]):
    return \
        [{'iteration': iteration_index,
          **iteration.iteration_statistics,
          'iteration_status': iteration.status.name,
          **_extract_model_info(iteration)}
         for iteration_index, iteration in enumerate(iterations)]


def extract_enriched_ngrams_and_weights_list(elements, boolean_labels):
    try:
        texts = [element.text for element in elements]
        enriched_ngrams_and_weights = ngrams_by_info_gain(texts, boolean_labels, ngram_max_length=2,
                                                          language=Languages.ENGLISH)
    except Exception as e:
        logging.warning(f"Failed to calculate enriched tokens from {len(elements)} elements: error {e}")
        enriched_ngrams_and_weights = {}

    ngrams_and_weights_list = [{'text': ngram, 'weight': weight} for ngram, weight in enriched_ngrams_and_weights]
    return ngrams_and_weights_list[:30]


def get_text_snippet(text, query_string, is_regex: bool):
    """
    Get snippet from full text.
    1. If length of text in tokens is less than snippet maximum length, the full text is returned.
    2. If there is no query, or query is not matched, the snippet is built from the start and the end of the text.
    3. if query is matched to the text, the snippet is built from up to 3 excerpts containing the query.
    Query must be contained by an exact match to the text (case in-sensitive).
    :param text:
    :param query_string:
    :return: a string containing snippet
    """

    text_tokens = text.split(" ")
    max_token_length = current_app.config['CONFIGURATION'].snippet_max_token_length
    if len(text_tokens) <= max_token_length:
        return text
    starting_text = " ".join(text_tokens[:int(max_token_length / 2)])
    end_text = " ".join(text_tokens[-int(max_token_length / 2):])
    if not query_string:
        return starting_text + " ... " + end_text
    
    query_string = re.escape(query_string)

    matches = [m for m in re.finditer(
        r"[\b]?[0-9a-zA-Z_`'.-]*" + query_string + r"[0-9a-zA-Z_`'.-]*[\b]?",
        text, 
        flags=re.IGNORECASE if is_regex is False else 0)][:3]

    if len(matches) == 0:
        return starting_text + " ... " + end_text
    
    snippet = ""
    num_tokens_before_after_match = max_token_length - len(matches) * len(query_string.split(" "))
    num_tokens_before_after_match = int(num_tokens_before_after_match/(2*len(matches)))
    for i, m in enumerate(matches):
        start_ind = m.regs[0][0]
        end_ind = m.regs[0][1]
        match = text[start_ind:end_ind]
        text_before_match = text[:start_ind].split()
        text_after_match = text[end_ind:].split()
        if i == 0 and start_ind - len(" ".join(text_before_match[-num_tokens_before_after_match:])) > 1:
            snippet += " ... "
        snippet += " ".join(text_before_match[-num_tokens_before_after_match:])
        snippet += " " + match + " "
        snippet += " ".join(text_after_match[:num_tokens_before_after_match])
        if end_ind + len(" ".join(text_after_match[:num_tokens_before_after_match])) < len(text):
            snippet += " ... "

    return snippet


def get_natural_sort_key(text):
    return [int(x) if x.isdigit() else x for x in re.split(r'(\d+)', text)]


def get_default_customizable_UI_text():
    """
    Returns a dict with the default values of the UI customizable elements
    """
    with open(os.path.join(os.path.dirname(__file__), "ui_defaults.json"), "rb") as f:
        return json.load(f)


def get_customizable_UI_text(path: str = None):
    """
    Returns a dict with the values of the UI customizable components, such as texts, urls and app logo.
    If path is None, the defaults are returned. If path is not None, the values in the file are
    returned. If some values are not present in the provided files, the default values are used.
    """
    default_customizable_UI_text = get_default_customizable_UI_text()
    if path is not None:
        try:
            with open(path, "rb") as f:
                customizable_UI_text = json.load(f)
        except FileNotFoundError as e:
            logging.error(f"The custom UI file was not found at {path}")
            return make_error({
                "type": "ui_customizable_elements_file_not_found",
                "title": "The json file with the UI customizable elements was not found.", 
            }, 404)
        wrong_keys = [k for k in customizable_UI_text.keys() if k not in default_customizable_UI_text.keys()]
        if len(wrong_keys) > 0:
            return make_error({
                "type": "wrong_customizable_keys",
                "title": f"The following keys in the provided customizable UI elements are not supported: "
                         f"{', '.join(wrong_keys)}."
            }, 404)
        merged = {dk: dv if dk not in customizable_UI_text else customizable_UI_text[dk]
                  for dk, dv in default_customizable_UI_text.items()}
        return merged
    else:
        return default_customizable_UI_text
    



#
#  Copyright (c) 2022 IBM Corp.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

import io
import os
import time
import tempfile
import unittest
from label_sleuth import app, config, app_utils
from label_sleuth.orchestrator.core.state_api.orchestrator_state_api import IterationStatus
from unittest import mock
import json

HEADERS = {'Content-Type': 'application/json'}


class TestAppIntegration(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.temp_dir = tempfile.TemporaryDirectory()
        path_to_test_config = os.path.abspath(os.path.join(__file__, os.pardir, "config_for_tests.json"))
        loaded_config = config.load_config(path_to_test_config)
        # avoid downloading spacy model for test
        loaded_config.language.spacy_model_name = None
        
        app_for_test = app.create_app(
            config=loaded_config,
            output_dir=os.path.join(cls.temp_dir.name, 'output'),
            customizable_ui_text_path=os.path.abspath(os.path.join(__file__, os.pardir, "ui_customs_for_tests.json"))
        )   
        app_for_test.test_request_context("/")
        app_for_test.config['TESTING'] = True
        app_for_test.config['LOGIN_DISABLED'] = True

        print(f"Integration tests, all output files will be written under {cls.temp_dir}")
        cls.client = app_for_test.test_client()

    @classmethod
    def tearDownClass(cls):
        cls.temp_dir.cleanup()

    def test_full_flow(self):

        ui_defaults = app_utils.get_default_customizable_UI_text()

        def raise_exception(a, b):
            # to make getting the customizable text json to fail and the defaults to work
            raise FileNotFoundError()
        
        # patching get_default_customizable_UI_text because it uses open() too
        with mock.patch('label_sleuth.app_utils.get_default_customizable_UI_text', side_effect=lambda: ui_defaults):
            with mock.patch('builtins.open', raise_exception):
                res = self.client.get("/customizable_ui_text", headers=HEADERS) 
        self.assertEqual(404, res.status_code)
        self.assertEqual(res.get_json()["title"], "The json file with the UI customizable elements was not found.")

        # use an incorrect key
        with mock.patch('label_sleuth.app_utils.get_default_customizable_UI_text', side_effect=lambda: ui_defaults):
            with mock.patch('builtins.open', mock.mock_open(read_data='{"this_key_doesnt_exist": "hello tests", "this_key_doesnt_exist_either": "hello tests"}')):
                res = self.client.get("/customizable_ui_text", headers=HEADERS)
        self.assertEqual(404, res.status_code)
        self.assertEqual(res.get_json()["title"], "The following keys in the provided customizable UI elements are not supported: this_key_doesnt_exist, this_key_doesnt_exist_either.")

        res = self.client.get("/customizable_ui_text", headers=HEADERS)
        correct_body = {
            "category_description_placeholder": "",
            "category_modal_helper_text": "Please select a meaningful name for your category.",
            "slack_link_url": "https://join.slack.com/t/labelsleuth/shared_invite/zt-1j5tpz1jl-W~UaNEKmK0RtzK~lI3Wkxg",
            "slack_link_title": "Join Slack",
            "github_link_url": "https://github.com/label-sleuth/label-sleuth",
            "github_link_title": "Github",
            "webpage_link_url": "https://www.label-sleuth.org/docs/index.html",
            "webpage_link_title": "Documentation",
            "ls_brief_description": "Quickly create a text classifier",
            "app_logo_path": "assets/sleuth_logo_white.svg",
            "document_upload_helper_text": "The csv file must have a header line (of \"text\" and optional \"document_id\")",
            "system_unavailable": "This is customized",
            "next_zero_shot_model_training_msg": "",
            "download_model_description": "",
            "download_model_bullets": [
                "the model itself", 
                "a code snippet demonstrating how it can be used within a Python application"
            ]
        }
        self.assertEqual(res.get_json(), correct_body)

        res = self.client.get("/assets/sleuth_logo_white.svg")
        self.assertEqual("image/svg+xml; charset=utf-8", res.content_type, msg="Failed to get app logo, request is returning index.html")

        self.maxDiff = None
        dataset_name = "my_test_dataset"
        workspace_name = "my_test_workspace"
        category_name = "my_category"
        category_description = "my_category_description"
        data = {}
        text_with_parenthesis = "this text contains a parenthesis a a a a a a(b b b b b b c c c c ( x x x x and some more text to force creating a snippet"
        text_with_parenthesis_snippet = "this text contains a parenthesis a a a a a ... x and some more text to force creating a snippet"
        text_with_parenthesis_snippet_in_query = " ... a a a a a(b b b b b ... c c c c ( x x x x ... "
        data['file'] = (io.BytesIO(bytes(
            'document_id,text\n' 
            'document1,this is the first text element of document one\n'
            'document1,this is the second text element of document one\n'
            'document2,this is the only text element in document two\n'
            'document3,"document 3 has three text elements, this is the first"\n'
            'document3,"document 3 has three text elements, this is the second that will be labeled as negative"\n'
            'document3,"document 3 has three text elements, this is the third"\n'
            f'document3,{text_with_parenthesis}\n', 'utf-8')),
                        'my_file.csv')
        res = self.client.post(f"/datasets/{dataset_name}/add_documents", data=data, headers=HEADERS,
                               content_type='multipart/form-data')
        self.assertEqual(200, res.status_code, msg="Failed to upload a new dataset")
        self.assertEqual(
            {'dataset_name': 'my_test_dataset', 'num_docs': 3, 'num_sentences': 7, 'workspaces_to_update': []},
            res.get_json(), msg="diff in upload dataset response")
        
        data['file'] = (io.BytesIO(
            b'document_id,text\n'
            b'document3,"adding this text element makes the request to fail because of the row count limit of 7"\n'),
                        'my_file.csv')
        res = self.client.post(f"/datasets/{dataset_name}/add_documents", data=data, headers=HEADERS,
                               content_type='multipart/form-data')
        
        self.assertEqual(409, res.status_code, msg="Failed to upload a new dataset")

        res = self.client.post("/workspace",
                               data='{{"workspace_id":"{}","dataset_id":"{}"}}'.format(workspace_name, dataset_name),
                               headers=HEADERS)
        self.assertEqual(200, res.status_code, msg="Failed to create a workspace")
        self.assertEqual(
            {"workspace": {'dataset_name': 'my_test_dataset', 'first_document_id': 'my_test_dataset-document1',
                           'workspace_id': 'my_test_workspace'}}, res.get_json(),
            msg="diff in create workspace response")
        res = self.client.post(f"/workspace/{workspace_name}/category",
                               data='{{"category_name":"{}","category_description":"{}"}}'.format(category_name,
                                                                                                  category_description),
                               headers=HEADERS)
        category_id = int(res.get_json()["category_id"])
        self.assertEqual(200, res.status_code, msg="Failed to add a new category to workspace")
        self.assertEqual({'category_description': 'my_category_description',
                          'category_name': category_name, 'category_id': str(category_id)}, res.get_json(),
                         msg="diff in create category response")

        res = self.client.get(f"/workspace/{workspace_name}/documents", headers=HEADERS)
        self.assertEqual(200, res.status_code, msg="Failed to get all documents uris")
        documents = res.get_json()['documents']
        self.assertEqual(3, len(documents),
                         msg="Number of retrieved documents is different the number of documents loaded")
        self.assertEqual({'documents': [{'document_id': 'my_test_dataset-document1'},
                                        {'document_id': 'my_test_dataset-document2'},
                                        {'document_id': 'my_test_dataset-document3'}]}, res.get_json(),
                         msg="diff in get documents")
        res = self.client.get(f"/workspace/{workspace_name}/document/{documents[-1]['document_id']}", headers=HEADERS)
        self.assertEqual(200, res.status_code, msg="Failed to add a document before labeling")
        document3_elements = res.get_json()['elements']

        self.assertEqual([
            {'begin': 0, 'docid': 'my_test_dataset-document3', 'end': 53, 'id': 'my_test_dataset-document3-0',
             'model_predictions': {}, 'text': 'document 3 has three text elements, this is the first',
             'user_labels': {}},
            {'begin': 54, 'docid': 'my_test_dataset-document3', 'end': 141, 'id': 'my_test_dataset-document3-1',
             'model_predictions': {},
             'text': 'document 3 has three text elements, this is the second that will be labeled as negative',
             'user_labels': {}},
            {'begin': 142, 'docid': 'my_test_dataset-document3', 'end': 195, 'id': 'my_test_dataset-document3-2',
             'model_predictions': {}, 'text': 'document 3 has three text elements, this is the third',
             'user_labels': {}},
            {
            "begin": 196, "docid": "my_test_dataset-document3", "end": 317, "id": "my_test_dataset-document3-3",
            "model_predictions": {}, "text": text_with_parenthesis,
            "user_labels": {}}], 
            document3_elements, msg=f"diff in {documents[-1]['document_id']} content")
        res = self.client.put(f'/workspace/{workspace_name}/element/{document3_elements[0]["id"]}',
                              data='{{"category_id":"{}","value":"{}"}}'.format(category_id, True), headers=HEADERS)
        self.assertEqual(200, res.status_code, msg="Failed to set the first label for a category")
        self.assertEqual({'category_id': str(category_id),
                          'element': {'begin': 0, 'docid': 'my_test_dataset-document3', 'end': 53,
                                      'id': 'my_test_dataset-document3-0', 'model_predictions': {},
                                      'text': 'document 3 has three text elements, this is the first',
                                      'user_labels': {str(category_id): 'true'}}, 'workspace_id': 'my_test_workspace'},
                         res.get_json(), msg="diff in setting element's label response")
        res = self.client.get(f"/workspace/{workspace_name}/status?category_id={category_id}",
                              headers=HEADERS)
        self.assertEqual(200, res.status_code, msg="Failed to get status after successfully setting the first label")
        self.assertEqual({'labeling_counts': {'true': 1}, 'progress': {'all': 33}},
                         res.get_json(), msg="diffs in get status response after setting a label")

        res = self.client.put(f'/workspace/{workspace_name}/element/{document3_elements[2]["id"]}',
                              data='{{"category_id":"{}","value":"{}"}}'.format(category_id, True), headers=HEADERS)

        self.assertEqual(200, res.status_code, msg="Failed to set the third label for category")
        self.assertEqual({'category_id': str(category_id),
                          'element': {'begin': 142, 'docid': 'my_test_dataset-document3', 'end': 195,
                                      'id': 'my_test_dataset-document3-2', 'model_predictions': {},
                                      'text': 'document 3 has three text elements, this is the third',
                                      'user_labels': {str(category_id): 'true'}}, 'workspace_id': 'my_test_workspace'},
                         res.get_json())

        res = self.client.get(f"/workspace/{workspace_name}/query?category_id={category_id}&qry_string=(",
                              headers=HEADERS)
        self.assertEqual(
            {"elements": [
                {
                    "begin": 196,"docid": "my_test_dataset-document3","end": 317,"id": "my_test_dataset-document3-3","model_predictions": {},
                    "snippet": text_with_parenthesis_snippet_in_query,
                    "text": text_with_parenthesis,
                    "user_labels": {}
                }
            ], "hit_count": 1,  "hit_count_unique": 1}, 
            res.get_json(), 
            msg="The searched text differs from the response"
        )
        res = self.client.get(f"/workspace/{workspace_name}/status?category_id={category_id}",
                              headers=HEADERS)
        self.assertEqual(200, res.status_code, msg="Failed to get status after successfully setting the second label")
        self.assertEqual({'labeling_counts': {'true': 2}, 'progress': {'all': 50}},
                         res.get_json(),
                         msg="diffs in get status response after setting the second label")


        res = self.client.put(f'/workspace/{workspace_name}/element/{document3_elements[1]["id"]}',
                              data='{{"category_id":"{}","value":"{}"}}'.format(category_id, False),
                              headers=HEADERS)
        self.assertEqual(200, res.status_code, msg="Failed to set the second label for a category")
        self.assertEqual({'category_id': str(category_id),
                          'element': {'begin': 54, 'docid': 'my_test_dataset-document3', 'end': 141,
                                      'id': 'my_test_dataset-document3-1', 'model_predictions': {},
                                      'text': 'document 3 has three text elements, '
                                              'this is the second that will be labeled as negative',
                                      'user_labels': {str(category_id): 'false'}}, 'workspace_id': 'my_test_workspace'},
                         res.get_json(), msg="diff in setting element's label response")
        res = self.client.get(f"/workspace/{workspace_name}/status?category_id={category_id}",
                              headers=HEADERS)
        self.assertEqual(200, res.status_code, msg="Failed to get status after successfully setting the third label")
        self.assertEqual({'true': 2, 'false': 1},
                         res.get_json()['labeling_counts'], msg="diffs in get status response after setting a label")

        res = self.wait_for_new_iteration(category_id, res, workspace_name, 1)

        self.assertEqual(200, res.status_code, msg="Failed to get iterations list")
        self.assertEqual(1, len(res.get_json()["iterations"]), msg="first model was not added to the models list")

        # get active learning recommendations
        res = self.client.get(f"/workspace/{workspace_name}/active_learning?category_id={category_id}",
                              headers=HEADERS)
        self.assertEqual(200, res.status_code, msg="Failed to get active learning recommendations")
        active_learning_response = res.get_json()
        self.assertEqual({'elements': [
            {'begin': 47, 'docid': 'my_test_dataset-document1', 'end': 94, 'id': 'my_test_dataset-document1-1',
             'model_predictions': {str(category_id): 'true'}, 'text': 'this is the second text element of document one',
             'user_labels': {}},
             {
             "begin": 196, "docid": "my_test_dataset-document3", "end": 317, "id": "my_test_dataset-document3-3",
             "model_predictions": {"0": "true"}, "snippet": text_with_parenthesis_snippet, "text": text_with_parenthesis,
             "user_labels": {}},
            {'begin': 0, 'docid': 'my_test_dataset-document2', 'end': 45, 'id': 'my_test_dataset-document2-0',
             'model_predictions': {str(category_id): 'true'}, 'text': 'this is the only text element in document two',
             'user_labels': {}},
            {'begin': 0, 'docid': 'my_test_dataset-document1', 'end': 46, 'id': 'my_test_dataset-document1-0',
             'model_predictions': {str(category_id): 'true'}, 'text': 'this is the first text element of document one',
             'user_labels': {}}], 'hit_count': 4},
            active_learning_response)

        # set the first label according to the active learning recommendations
        res = self.client.put(f'/workspace/{workspace_name}/element/{active_learning_response["elements"][0]["id"]}',
                              data='{{"category_id":"{}","value":"{}"}}'.format(category_id, True), headers=HEADERS)

        self.assertEqual(200, res.status_code,
                         msg="Failed to set the label for the first element recommended by the active learning")
        self.assertEqual({'category_id': str(category_id),
                          'element': {'begin': 47, 'docid': 'my_test_dataset-document1', 'end': 94,
                                      'id': 'my_test_dataset-document1-1',
                                      'model_predictions': {str(category_id): 'true'},
                                      'text': 'this is the second text element of document one',
                                      'user_labels': {str(category_id): 'true'}}, 'workspace_id': 'my_test_workspace'},
                         res.get_json())

        res = self.client.get(f"/workspace/{workspace_name}/status?category_id={category_id}",
                              headers=HEADERS)
        self.assertEqual(200, res.status_code,
                         msg="Failed to get status after successfully setting the first label for the second model")
        self.assertEqual({'labeling_counts': {'true': 3, 'false': 1}, 'progress': {'all': 33}},
                         res.get_json(), msg="diffs in get status response after setting a label")

        # set the second label according to the active learning recommendations
        res = self.client.put(f'/workspace/{workspace_name}/element/{active_learning_response["elements"][2]["id"]}',
                              data='{{"category_id":"{}","value":"{}"}}'.format(category_id, False),
                              headers=HEADERS)

        self.assertEqual(200, res.status_code,
                         msg="Failed to set the label for the first element recommended by the active learning")
        self.assertEqual({'category_id': str(category_id),
                          'element': {'begin': 0, 'docid': 'my_test_dataset-document2', 'end': 45,
                                      'id': 'my_test_dataset-document2-0',
                                      'model_predictions': {str(category_id): 'true'},
                                      'text': 'this is the only text element in document two',
                                      'user_labels': {str(category_id): 'false'}}, 'workspace_id': 'my_test_workspace'},
                         res.get_json())

        res = self.client.get(f"/workspace/{workspace_name}/status?category_id={category_id}",
                              headers=HEADERS)
        self.assertEqual(200, res.status_code,
                         msg="Failed to get status after successfully setting the second label for the second model")
        self.assertEqual({'true': 3, 'false': 2},
                         res.get_json()['labeling_counts'], msg="diffs in get status response after setting a label")

        # get positively labeled elements
        res = self.client.get(f"/workspace/{workspace_name}/positive_elements?category_id={category_id}",
                              headers=HEADERS)
        self.assertEqual(200, res.status_code,
                         msg="Failed to get positively labeled elements")
        self.assertEqual({'elements': [
                            {'begin': 0,
                            'docid': 'my_test_dataset-document3',
                            'end': 53,
                            'id': 'my_test_dataset-document3-0',
                            'model_predictions': {'0': 'true'},
                            'text': 'document 3 has three text elements, this is '
                                    'the first',
                            'user_labels': {'0': 'true'}},
                            {'begin': 47,
                            'docid': 'my_test_dataset-document1',
                            'end': 94,
                            'id': 'my_test_dataset-document1-1',
                            'model_predictions': {'0': 'true'},
                            'text': 'this is the second text element of document '
                                    'one',
                            'user_labels': {'0': 'true'}},
                            {'begin': 142,
                            'docid': 'my_test_dataset-document3',
                            'end': 195,
                            'id': 'my_test_dataset-document3-2',
                            'model_predictions': {'0': 'true'},
                            'text': 'document 3 has three text elements, this is '
                                    'the third',
                            'user_labels': {'0': 'true'}}], 
                        'hit_count': 3},
                        res.get_json(), msg="diffs in positively labeled elements")


        # get negatively labeled elements
        res = self.client.get(f"/workspace/{workspace_name}/negative_elements?category_id={category_id}",
                              headers=HEADERS)
        self.assertEqual(200, res.status_code,
                         msg="Failed to get negatively labeled elements")
        self.assertEqual({'elements': [
                            {'begin': 0,
                            'docid': 'my_test_dataset-document2',
                            'end': 45,
                            'id': 'my_test_dataset-document2-0',
                            'model_predictions': {'0': 'true'},
                            'text': 'this is the only text element in document two',
                            'user_labels': {'0': 'false'}},
                            {'begin': 54,
                            'docid': 'my_test_dataset-document3',
                            'end': 141,
                            'id': 'my_test_dataset-document3-1',
                            'model_predictions': {'0': 'false'},
                            'text': 'document 3 has three text elements, this is '
                                    'the second that will be labeled as negative',
                            'user_labels': {'0': 'false'}}],
                        'hit_count': 2},
                        res.get_json(), msg="diffs in negatively labeled elements")


        # continue labeling from AL suggestions to get a new model (removing a negative label -> false count decrease)
        res = self.client.put(f'/workspace/{workspace_name}/element/{active_learning_response["elements"][2]["id"]}',
                              data='{{"category_id":"{}","value":"{}"}}'.format(category_id, True), headers=HEADERS)
        self.assertEqual(200, res.status_code,
                         msg="Failed to set the label for the first element recommended by the active learning")
        self.assertEqual({'category_id': '0', 
                          'element': {
                              'begin': 0, 'docid': 'my_test_dataset-document2', 'end': 45, 'id': 'my_test_dataset-document2-0', 
                              'model_predictions': {'0': 'true'}, 'text': 'this is the only text element in document two', 
                              'user_labels': {'0': 'true'}}, 'workspace_id': 'my_test_workspace'},
                         res.get_json())

        res = self.client.get(f"/workspace/{workspace_name}/status?category_id={category_id}",
                              headers=HEADERS)
        self.assertEqual(200, res.status_code,
                         msg="Failed to get status after successfully setting the third label for the second model")
        self.assertEqual({'true': 4, 'false': 1},
                         res.get_json()['labeling_counts'], msg="diffs in get status response after setting a label")


        # wait for the second models
        res = self.wait_for_new_iteration(category_id, res, workspace_name, 2)
        self.assertEqual(200, res.status_code, msg="Failed to get models list")
        self.assertEqual(2, len(res.get_json()["iterations"]), msg="second model was not added to the models list")

        # update existing category
        new_category_name = f'{category_name}_new'
        new_category_description = f'{category_description} new'
        res = self.client.put(f"/workspace/{workspace_name}/category/{category_id}",
                              data='{{"category_name":"{}","category_description":"{}"}}'.format(
                                  new_category_name, new_category_description),
                              headers=HEADERS)
        self.assertEqual(200, res.status_code, msg="Failed to update category")

        # delete category and create a new one
        res = self.client.delete(f"/workspace/{workspace_name}/category/{category_id}", headers=HEADERS)
        self.assertEqual(200, res.status_code, msg="Failed to delete category")

        res = self.client.get(f"/workspace/{workspace_name}/categories", headers=HEADERS)
        self.assertEqual(len(res.get_json()['categories']), 0, msg='unexpected number of categories')

        res = self.client.post(f"/workspace/{workspace_name}/category",
                               data='{{"category_name":"{}","category_description":"{}"}}'.format(category_name,
                                                                                                  category_description),
                               headers=HEADERS)
        category_id = int(res.get_json()["category_id"])
        self.assertEqual(200, res.status_code, msg="Failed to add a new category to workspace")
        self.assertEqual(res.get_json(),
                         {'category_description': category_description, 'category_name': category_name,
                          'category_id': str(category_id)}, msg="diff in create category response")

        res = self.client.get(f"/workspace/{workspace_name}/categories", headers=HEADERS)
        self.assertEqual(len(res.get_json()['categories']), 1, msg='unexpected number of categories')

        print("Done")

    def wait_for_new_iteration(self, category_id, res, workspace_name, num_models):
        waiting_count = 0
        MAX_WAITING_FOR_TRAINING = 100
        # wait maximum 10 seconds (100*0.1) for the training
        # (should be much faster but it sometimes takes longer on github actions)
        while waiting_count < MAX_WAITING_FOR_TRAINING:
            # since get_status is asynchronously starting a new training, we need to wait until it added to the
            # iterations list and finishes successfully
            res = self.client.get(f"/workspace/{workspace_name}/iterations?category_id={category_id}",
                                  headers=HEADERS)
            response = res.get_json()

            if res.status_code != 200 or \
                    (len(response["iterations"]) == num_models
                     and response['iterations'][num_models-1]['iteration_status'] == IterationStatus.READY.name):
                break
            time.sleep(0.1)
            waiting_count += 1
        return res



#
#  Copyright (c) 2022 IBM Corp.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

from dataclasses import dataclass
from typing import Sequence


@dataclass
class Language:
    """
    This dataclass contains some parameters that are language-specific. The parameter values can be used by the
    classification model implementations, as well as by some general utility functions in the *analysis_utils* module.
    """
    name: str
    stop_words: Sequence[str]
    spacy_model_name: str = None
    fasttext_language_id: str = None
    right_to_left: bool = False

    def __repr__(self):
        return self.name

    def __hash__(self):
        return hash(self.name)

    def __post_init__(self):
        if self.spacy_model_name is None and self.fasttext_language_id is None:
            raise Exception(f"Selected Language {self.name} should define either "
                            f"spacy_model_name or fasttext_language_id in class '{Language}'")


English = \
    Language(name='English',
             stop_words=["a", "about", "above", "after", "again", "against", "ain", "all", "am", "an", "and", "any",
                         "are", "aren", "aren't", "as", "at", "be", "because", "been", "before", "being", "below",
                         "between", "both", "but", "by", "can", "couldn", "couldn't", "d", "did", "didn", "didn't",
                         "do", "does", "doesn", "doesn't", "doing", "don", "don't", "down", "during", "each", "few",
                         "for", "from", "further", "had", "hadn", "hadn't", "has", "hasn", "hasn't", "have", "haven",
                         "haven't", "having", "he", "her", "here", "hers", "herself", "him", "himself", "his", "how",
                         "i", "if", "in", "into", "is", "isn", "isn't", "it", "it's", "its", "itself", "just", "ll",
                         "m", "ma", "me", "mightn", "mightn't", "more", "most", "mustn", "mustn't", "my", "myself",
                         "needn", "needn't", "no", "nor", "not", "now", "o", "of", "off", "on", "once", "only", "or",
                         "other", "our", "ours", "ourselves", "out", "over", "own", "re", "s", "same", "shan", "shan't",
                         "she", "she's", "should", "should've", "shouldn", "shouldn't", "so", "some", "such", "t",
                         "than", "that", "that'll", "the", "their", "theirs", "them", "themselves", "then", "there",
                         "these", "they", "this", "those", "through", "to", "too", "under", "until", "up", "ve", "very",
                         "was", "wasn", "wasn't", "we", "were", "weren", "weren't", "what", "when", "where", "which",
                         "while", "who", "whom", "why", "will", "with", "won", "won't", "wouldn", "wouldn't", "y",
                         "you", "you'd", "you'll", "you're", "you've", "your", "yours", "yourself", "yourselves",
                         "could", "he'd", "he'll", "he's", "here's", "how's", "i'd", "i'll", "i'm", "i've", "let's",
                         "ought", "she'd", "she'll", "that's", "there's", "they'd", "they'll", "they're", "they've",
                         "we'd", "we'll", "we're", "we've", "what's", "when's", "where's", "who's", "why's", "would"],
             spacy_model_name='en_core_web_lg')

Italian = \
    Language(name='Italian',
             stop_words=["scola", "varie", "saro", "stanno", "dagli", "certa", "le", "nemmeno", "molto", "avuti",
                         "posteriore", "facesse", "il", "nostra", "durante", "faceste", "dalla", "può", "della", "ora",
                         "faccia", "alla", "dice", "scorso", "probabilmente", "volta", "sareste", "tre", "fu", "dell",
                         "quanta", "soltanto", "fa", "degl", "diventare", "pure", "quantunque", "che", "anche", "pieno",
                         "nostri", "te", "facessero", "qualunque", "codesti", "nonsia", "come", "successivamente",
                         "stato", "mesi", "quelle", "insieme", "agli", "volte", "faresti", "nell", "lui", "ma",
                         "malgrado", "ossia", "allo", "lato", "mi", "aveste", "lo", "molti", "basta", "seguito",
                         "dallo", "cioe", "giacche", "avevate", "starebbero", "parecchio", "ansa", "paese", "partendo",
                         "secondo", "prima", "gliela", "quale", "là", "uno", "vari", "facciamo", "dove", "sarà",
                         "stati", "medesimo", "steste", "male", "stiate", "facciano", "ahime", "dovrà", "tra",
                         "stessero", "osi", "esse", "seguente", "stavo", "percio", "sarebbe", "avrai", "parecchie",
                         "codesto", "dall", "cominci", "tempo", "sui", "relativo", "faremo", "aver", "staremo", "citta",
                         "dei", "improvviso", "stesse", "verso", "generale", "po", "avreste", "mie", "parecchi",
                         "fuori", "uguali", "gruppo", "queste", "contro", "diventato", "col", "così", "puo", "sarete",
                         "quindi", "anni", "si", "stavamo", "mediante", "eravamo", "state", "stettero", "sto", "colui",
                         "facesti", "lungo", "altro", "fareste", "fummo", "agl", "solito", "certe", "abbia", "ai",
                         "faremmo", "stavano", "perfino", "no", "quelli", "certo", "miei", "feci", "fossi", "presa",
                         "purtroppo", "con", "facevi", "chi", "momento", "concernente", "stiamo", "ecc", "saremo",
                         "successivo", "una", "per", "coloro", "governo", "nessuno", "dappertutto", "conclusione",
                         "fra", "farà", "glieli", "sotto", "salvo", "dentro", "altri", "sarebbero", "fece", "tutto",
                         "quel", "effettivamente", "luogo", "alle", "essi", "ti", "ella", "sembrare", "casa", "foste",
                         "fatto", "sarò", "starò", "la", "avuta", "tu", "codesta", "in", "sugl", "staresti", "starà",
                         "fanno", "starete", "mila", "degli", "stesti", "avute", "voi", "esser", "finalmente", "avemmo",
                         "ebbe", "peggio", "cosa", "nondimeno", "maggior", "lei", "grande", "bene", "avevo", "ore",
                         "stava", "nello", "stiano", "allora", "farò", "infatti", "sembrato", "dirimpetto", "nella",
                         "me", "ero", "piglia", "altrove", "nel", "saremmo", "ho", "entrambi", "ognuno", "altrimenti",
                         "tanto", "avranno", "dal", "recentemente", "tali", "sulle", "tua", "dovunque", "farebbe",
                         "sei", "avessimo", "pero", "avremmo", "coll", "facevate", "spesso", "avevano", "cio", "niente",
                         "inoltre", "ancora", "otto", "li", "abbastanza", "varia", "qualcuno", "vale", "già", "ieri",
                         "vostra", "cento", "facemmo", "stette", "faceva", "persone", "avevamo", "ci", "troppo", "però",
                         "dunque", "sulla", "avrete", "all", "vi", "piuttosto", "saresti", "stessi", "ottanta", "bravo",
                         "sia", "ed", "proprio", "intanto", "quello", "nelle", "chiunque", "facevano", "visto", "colei",
                         "miliardi", "avrei", "press", "invece", "nulla", "ahimè", "altrui", "se", "sono", "sua",
                         "gliele", "consiglio", "ulteriore", "peccato", "recente", "tuoi", "lavoro", "frattempo",
                         "avrà", "qualcosa", "eppure", "sopra", "stando", "forza", "fossero", "attraverso",
                         "nonostante", "gliene", "dai", "negli", "farete", "moltissimo", "poiche", "sempre", "quante",
                         "noi", "sembri", "dietro", "marche", "poco", "loro", "nei", "vostre", "giorno", "ad", "città",
                         "deve", "novanta", "oltre", "eri", "grazie", "milioni", "quest", "talvolta", "cui", "glielo",
                         "dalle", "lasciato", "certi", "facessi", "solo", "avrebbero", "essendo", "quanti", "stia",
                         "hai", "stemmo", "via", "dov", "staremmo", "avente", "nove", "perciò", "quella", "mai", "tuo",
                         "nostre", "lontano", "ex", "nostro", "ciascuno", "fino", "perché", "ultimo", "faranno", "ha",
                         "inc", "perche", "al", "ognuna", "sara", "modo", "ciascuna", "faccio", "qui", "chicchessia",
                         "cos", "mentre", "adesso", "sig", "hanno", "cosi", "avrò", "persino", "dovra", "avresti",
                         "facendo", "favore", "benissimo", "stareste", "sul", "da", "stata", "co", "quanto", "tale",
                         "facessimo", "avanti", "dagl", "tuttavia", "farai", "quando", "avendo", "preferibilmente",
                         "avremo", "starebbe", "finche", "caso", "stetti", "malissimo", "essere", "giorni", "stesso",
                         "vostri", "vario", "trenta", "ecco", "fosti", "stai", "ne", "mia", "furono", "vita", "sarei",
                         "poi", "anticipo", "delle", "tranne", "principalmente", "coi", "stavate", "potrebbe", "farei",
                         "siete", "brava", "fin", "cima", "attesa", "stavi", "del", "fecero", "avesse", "forse",
                         "detto", "gia", "farebbero", "avessero", "saranno", "minimi", "piu", "senza", "starei",
                         "sarai", "possedere", "non", "pochissimo", "un", "primo", "piedi", "davanti", "titolo", "od",
                         "uomo", "stessimo", "mio", "negl", "tutte", "avrebbe", "suo", "fossimo", "fai", "io", "mosto",
                         "parte", "facevo", "questa", "facevamo", "tutta", "sette", "gli", "fui", "scopo", "vostro",
                         "fine", "fosse", "trovato", "nessuna", "qualcuna", "riecco", "cogli", "sta", "possa", "ebbero",
                         "su", "avete", "fare", "nuovo", "posto", "a", "realmente", "meno", "registrazione", "quasi",
                         "mondo", "sullo", "comunque", "siano", "dopo", "esempio", "avevi", "siamo", "avesti", "nessun",
                         "tue", "va", "cortesia", "intorno", "questi", "ministro", "anno", "alcuni", "sugli", "tutti",
                         "srl", "diventa", "subito", "nazionale", "siate", "ebbi", "sue", "alcuna", "abbiate", "mezzo",
                         "due", "meglio", "eravate", "ogni", "macche", "dello", "magari", "più", "alcuno", "questo",
                         "avere", "oggi", "vicino", "futuro", "circa", "staranno", "stessa", "di", "avuto", "haha",
                         "assai", "starai", "abbiamo", "era", "oppure", "quali", "accidenti", "abbiano", "suoi", "egli",
                         "neppure", "averlo", "qualche", "sull", "avessi", "quattro", "facciate", "dire", "mancanza",
                         "affinche", "erano", "conciliarsi", "sembra", "aveva"],
             fasttext_language_id='it')

Romanian = \
    Language(name='Romanian',
             stop_words=["a", "abia", "acea", "aceasta", "această", "aceea", "aceeasi", "acei", "aceia", "acel",
                         "acela", "acelasi", "acele", "acelea", "acest", "acesta", "aceste", "acestea", "acestei",
                         "acestia", "acestui", "aceşti", "aceştia", "acolo", "acord", "acum", "adica", "ai", "aia",
                         "aibă", "aici", "aiurea", "al", "ala", "alaturi", "ale", "alea", "alt", "alta", "altceva",
                         "altcineva", "alte", "altfel", "alti", "altii", "altul", "am", "anume", "apoi", "ar", "are",
                         "as", "asa", "asemenea", "asta", "astazi", "astea", "astfel", "astăzi", "asupra", "atare",
                         "atat", "atata", "atatea", "atatia", "ati", "atit", "atita", "atitea", "atitia", "atunci",
                         "au", "avea", "avem", "aveţi", "avut", "azi", "aş", "aşadar", "aţi", "b", "ba", "bine",
                         "bucur", "bună", "c", "ca", "cam", "cand", "capat", "care", "careia", "carora", "caruia",
                         "cat", "catre", "caut", "ce", "cea", "ceea", "cei", "ceilalti", "cel", "cele", "celor", "ceva",
                         "chiar", "ci", "cinci", "cind", "cine", "cineva", "cit", "cita", "cite", "citeva", "citi",
                         "citiva", "conform", "contra", "cu", "cui", "cum", "cumva", "curând", "curînd", "când", "cât",
                         "câte", "câtva", "câţi", "cînd", "cît", "cîte", "cîtva", "cîţi", "că", "căci", "cărei",
                         "căror", "cărui", "către", "d", "da", "daca", "dacă", "dar", "dat", "datorită", "dată", "dau",
                         "de", "deasupra", "deci", "decit", "degraba", "deja", "deoarece", "departe", "desi", "despre",
                         "deşi", "din", "dinaintea", "dintr", "dintr-", "dintre", "doar", "doi", "doilea", "două",
                         "drept", "dupa", "după", "dă", "e", "ea", "ei", "el", "ele", "era", "eram", "este", "eu",
                         "exact", "eşti", "f", "face", "fara", "fata", "fel", "fi", "fie", "fiecare", "fii", "fim",
                         "fiu", "fiţi", "foarte", "fost", "frumos", "fără", "g", "geaba", "graţie", "h", "halbă", "i",
                         "ia", "iar", "ieri", "ii", "il", "imi", "in", "inainte", "inapoi", "inca", "incit", "insa",
                         "intr", "intre", "isi", "iti", "j", "k", "l", "la", "le", "li", "lor", "lui", "lângă", "lîngă",
                         "m", "ma", "mai", "mare", "mea", "mei", "mele", "mereu", "meu", "mi", "mie", "mine", "mod",
                         "mult", "multa", "multe", "multi", "multă", "mulţi", "mulţumesc", "mâine", "mîine", "mă", "n",
                         "ne", "nevoie", "ni", "nici", "niciodata", "nicăieri", "nimeni", "nimeri", "nimic", "niste",
                         "nişte", "noastre", "noastră", "noi", "noroc", "nostri", "nostru", "nou", "noua", "nouă",
                         "noştri", "nu", "numai", "o", "opt", "or", "ori", "oricare", "orice", "oricine", "oricum",
                         "oricând", "oricât", "oricînd", "oricît", "oriunde", "p", "pai", "parca", "patra", "patru",
                         "patrulea", "pe", "pentru", "peste", "pic", "pina", "plus", "poate", "pot", "prea", "prima",
                         "primul", "prin", "printr-", "putini", "puţin", "puţina", "puţină", "până", "pînă", "r", "rog",
                         "s", "sa", "sa-mi", "sa-ti", "sai", "sale", "sau", "se", "si", "sint", "sintem", "spate",
                         "spre", "sub", "sunt", "suntem", "sunteţi", "sus", "sută", "sînt", "sîntem", "sînteţi", "să",
                         "săi", "său", "t", "ta", "tale", "te", "ti", "timp", "tine", "toata", "toate", "toată",
                         "tocmai", "tot", "toti", "totul", "totusi", "totuşi", "toţi", "trei", "treia", "treilea", "tu",
                         "tuturor", "tăi", "tău", "u", "ul", "ului", "un", "una", "unde", "undeva", "unei", "uneia",
                         "unele", "uneori", "unii", "unor", "unora", "unu", "unui", "unuia", "unul", "v", "va", "vi",
                         "voastre", "voastră", "voi", "vom", "vor", "vostru", "vouă", "voştri", "vreme", "vreo",
                         "vreun", "vă", "x", "z", "zece", "zero", "zi", "zice", "îi", "îl", "îmi", "împotriva", "în",
                         "înainte", "înaintea", "încotro", "încât", "încît", "între", "întrucât", "întrucît", "îţi",
                         "ăla", "ălea", "ăsta", "ăstea", "ăştia", "şapte", "şase", "şi", "ştiu", "ţi", "ţie"],
             spacy_model_name='ro_core_news_lg')

Hebrew = Language(name='Hebrew',
                  stop_words=["זה", "זאת", "אלה", "אלו", "גם", "של", "מה", "או"],
                  fasttext_language_id="he",
                  right_to_left=True)

Arabic = Language(name='Arabic',
                  stop_words=["التى", "التي", "الذى", "الذي", "الذين", "ذلك", "هذا", "هذه", "هؤلاء", "قد", "وقد", "حيث",
                              "ان", "إن", "انه", "وان", "فان", "فإن", "بان", "اي", "أي", "ايضا", "أيضا", "إياه"],
                  fasttext_language_id='ar',
                  right_to_left=True)

Spanish = Language(name='Spanish',
                   stop_words=['de', 'la', 'que', 'el', 'en', 'y', 'a', 'los', 'del', 'se', 'las', 'por', 'un', 'para',
                               'con', 'no', 'una', 'su', 'al', 'lo', 'como', 'más', 'pero', 'sus', 'le', 'ya', 'o',
                               'este', 'sí', 'porque', 'esta', 'entre', 'cuando', 'muy', 'sin', 'sobre', 'también',
                               'me', 'hasta', 'hay', 'donde', 'quien', 'desde', 'todo', 'nos', 'durante', 'todos',
                               'uno', 'les', 'ni', 'contra', 'otros', 'ese', 'eso', 'ante', 'ellos', 'e', 'esto', 'mí',
                               'antes', 'algunos', 'qué', 'unos', 'yo', 'otro', 'otras', 'otra', 'él', 'tanto', 'esa',
                               'estos', 'mucho', 'quienes', 'nada', 'muchos', 'cual', 'poco', 'ella', 'estar', 'estas',
                               'algunas', 'algo', 'nosotros', 'mi', 'mis', 'tú', 'te', 'ti', 'tu', 'tus', 'ellas',
                               'nosotras', 'vosostros', 'vosostras', 'os', 'mío', 'mía', 'míos', 'mías', 'tuyo', 'tuya',
                               'tuyos', 'tuyas', 'suyo', 'suya', 'suyos', 'suyas', 'nuestro', 'nuestra', 'nuestros',
                               'nuestras', 'vuestro', 'vuestra', 'vuestros', 'vuestras', 'esos', 'esas', 'estoy',
                               'estás', 'está', 'estamos', 'estáis', 'están', 'esté', 'estés', 'estemos', 'estéis',
                               'estén', 'estaré', 'estarás', 'estará', 'estaremos', 'estaréis', 'estarán', 'estaría',
                               'estarías', 'estaríamos', 'estaríais', 'estarían', 'estaba', 'estabas', 'estábamos',
                               'estabais', 'estaban', 'estuve', 'estuviste', 'estuvo', 'estuvimos', 'estuvisteis',
                               'estuvieron', 'estuviera', 'estuvieras', 'estuviéramos', 'estuvierais', 'estuvieran',
                               'estuviese', 'estuvieses', 'estuviésemos', 'estuvieseis', 'estuviesen', 'estando',
                               'estado', 'estada', 'estados', 'estadas', 'estad', 'he', 'has', 'ha', 'hemos', 'habéis',
                               'han', 'haya', 'hayas', 'hayamos', 'hayáis', 'hayan', 'habré', 'habrás', 'habrá',
                               'habremos', 'habréis', 'habrán', 'habría', 'habrías', 'habríamos', 'habríais', 'habrían',
                               'había', 'habías', 'habíamos', 'habíais', 'habían', 'hube', 'hubiste', 'hubo', 'hubimos',
                               'hubisteis', 'hubieron', 'hubiera', 'hubieras', 'hubiéramos', 'hubierais', 'hubieran',
                               'hubiese', 'hubieses', 'hubiésemos', 'hubieseis', 'hubiesen', 'habiendo', 'habido',
                               'habida', 'habidos', 'habidas', 'soy', 'eres', 'es', 'somos', 'sois', 'son', 'sea',
                               'seas', 'seamos', 'seáis', 'sean', 'seré', 'serás', 'será', 'seremos', 'seréis', 'serán',
                               'sería', 'serías', 'seríamos', 'seríais', 'serían', 'era', 'eras', 'éramos', 'erais',
                               'eran', 'fui', 'fuiste', 'fue', 'fuimos', 'fuisteis', 'fueron', 'fuera', 'fueras',
                               'fuéramos', 'fuerais', 'fueran', 'fuese', 'fueses', 'fuésemos', 'fueseis', 'fuesen',
                               'tengo', 'tienes', 'tiene', 'tenemos', 'tenéis', 'tienen', 'tenga', 'tengas', 'tengamos',
                               'tengáis', 'tengan', 'tendré', 'tendrás', 'tendrá', 'tendremos', 'tendréis', 'tendrán',
                               'tendría', 'tendrías', 'tendríamos', 'tendríais', 'tendrían', 'tenía', 'tenías',
                               'teníamos', 'teníais', 'tenían', 'tuve', 'tuviste', 'tuvo', 'tuvimos', 'tuvisteis',
                               'tuvieron', 'tuviera', 'tuvieras', 'tuviéramos', 'tuvierais', 'tuvieran', 'tuviese',
                               'tuvieses', 'tuviésemos', 'tuvieseis', 'tuviesen', 'teniendo', 'tenido', 'tenida',
                               'tenidos', 'tenidas', 'tened'],
                   fasttext_language_id='es')

# languages with fasttext word vectors (from https://fasttext.cc/docs/en/crawl-vectors.html), and their associated ids
fasttext_language_name_to_id = \
    {"Afrikaans": "af", "Albanian": "sq", "Alemannic": "als", "Amharic": "am", "Arabic": "ar", "Aragonese": "an",
     "Armenian": "hy", "Assamese": "as", "Asturian": "ast", "Azerbaijani": "az", "Bashkir": "ba", "Basque": "eu",
     "Bavarian": "bar", "Belarusian": "be", "Bengali": "bn", "Bihari": "bh", "Bishnupriya Manipuri": "bpy",
     "Bosnian": "bs", "Breton": "br", "Bulgarian": "bg", "Burmese": "my", "Catalan": "ca", "Cebuano": "ceb",
     "Central Bicolano": "bcl", "Chechen": "ce", "Chinese": "zh", "Chuvash": "cv", "Corsican": "co", "Croatian": "hr",
     "Czech": "cs", "Danish": "da", "Divehi": "dv", "Dutch": "nl", "Eastern Punjabi": "pa", "Egyptian Arabic": "arz",
     "Emilian-Romagnol": "eml", "English": "en", "Erzya": "myv", "Esperanto": "eo", "Estonian": "et",
     "Fiji Hindi": "hif", "Finnish": "fi", "French": "fr", "Galician": "gl", "Georgian": "ka", "German": "de",
     "Goan Konkani": "gom", "Greek": "el", "Gujarati": "gu", "Haitian": "ht", "Hebrew": "he", "Hill Mari": "mrj",
     "Hindi": "hi", "Hungarian": "hu", "Icelandic": "is", "Ido": "io", "Ilokano": "ilo", "Indonesian": "id",
     "Interlingua": "ia", "Irish": "ga", "Italian": "it", "Japanese": "ja", "Javanese": "jv", "Kannada": "kn",
     "Kapampangan": "pam", "Kazakh": "kk", "Khmer": "km", "Kirghiz": "ky", "Korean": "ko", "Kurdish (Kurmanji)": "ku",
     "Kurdish (Sorani)": "ckb", "Latin": "la", "Latvian": "lv", "Limburgish": "li", "Lithuanian": "lt",
     "Lombard": "lmo", "Low Saxon": "nds", "Luxembourgish": "lb", "Macedonian": "mk", "Maithili": "mai",
     "Malagasy": "mg", "Malay": "ms", "Malayalam": "ml", "Maltese": "mt", "Manx": "gv", "Marathi": "mr",
     "Mazandarani": "mzn", "Meadow Mari": "mhr", "Minangkabau": "min", "Mingrelian": "xmf", "Mirandese": "mwl",
     "Mongolian": "mn", "Nahuatl": "nah", "Neapolitan": "nap", "Nepali": "ne", "Newar": "new", "North Frisian": "frr",
     "Northern Sotho": "nso", "Norwegian (Bokmål)": "no", "Norwegian (Nynorsk)": "nn", "Occitan": "oc", "Oriya": "or",
     "Ossetian": "os", "Palatinate German": "pfl", "Pashto": "ps", "Persian": "fa", "Piedmontese": "pms",
     "Polish": "pl", "Portuguese": "pt", "Quechua": "qu", "Romanian": "ro", "Romansh": "rm", "Russian": "ru",
     "Sakha": "sah", "Sanskrit": "sa", "Sardinian": "sc", "Scots": "sco", "Scottish Gaelic": "gd", "Serbian": "sr",
     "Serbo-Croatian": "sh", "Sicilian": "scn", "Sindhi": "sd", "Sinhalese": "si", "Slovak": "sk", "Slovenian": "sl",
     "Somali": "so", "Southern Azerbaijani": "azb", "Spanish": "es", "Sundanese": "su", "Swahili": "sw",
     "Swedish": "sv", "Tagalog": "tl", "Tajik": "tg", "Tamil": "ta", "Tatar": "tt", "Telugu": "te", "Thai": "th",
     "Tibetan": "bo", "Turkish": "tr", "Turkmen": "tk", "Ukrainian": "uk", "Upper Sorbian": "hsb", "Urdu": "ur",
     "Uyghur": "ug", "Uzbek": "uz", "Venetian": "vec", "Vietnamese": "vi", "Volapük": "vo", "Walloon": "wa",
     "Waray": "war", "Welsh": "cy", "West Flemish": "vls", "West Frisian": "fy", "Western Punjabi": "pnb",
     "Yiddish": "yi", "Yoruba": "yo", "Zazaki": "diq", "Zeelandic": "zea"}


def add_auto_languages(cls):
    for language_name, language_id in fasttext_language_name_to_id.items():
        if not hasattr(cls, language_name.upper()):
            auto_language = Language(name=language_name, stop_words=[], fasttext_language_id=language_id)
            setattr(cls, language_name.upper(), auto_language)
    return cls


@add_auto_languages
class Languages:
    """
    This class contains Language objects for all languages supported by Label Sleuth. A few Language objects are defined
    explicitly below - these languages are initialized with a list of stop-words and/or some specific attributes;
    ~150 additional Language objects are initialized automatically and added to the class by the *add_auto_languages*
    decorator. The full list of supported languages can be found in *fasttext_language_name_to_id*.
    """
    ENGLISH = English
    ITALIAN = Italian
    ROMANIAN = Romanian
    HEBREW = Hebrew
    ARABIC = Arabic
    SPANISH = Spanish

    @classmethod
    def all_languages(cls):
        return {value for name, value in vars(cls).items() if name.isupper()}



from typing import Sequence
from label_sleuth.active_learning.core.active_learning_api import ActiveLearner
from label_sleuth.data_access.core.data_structs import TextElement
from label_sleuth.models.core.prediction import Prediction
from operator import itemgetter


class CombinedHMRetro(ActiveLearner):
    """
    This active learning module combines "hard" examples, i.e. examples that the model is most uncertain about, 
    with positive confident examples, by selecting them alternately.
    """
    def get_per_element_score(self, candidate_text_elements: Sequence[TextElement],
                              candidate_text_element_predictions: Sequence[Prediction], workspace_id: str,
                              dataset_name: str, category_id: int) -> Sequence[float]:
        hm_scores = [2*(0.5-abs(pred.score-0.5)) for pred in candidate_text_element_predictions]
        hm_indices, hm_sorted = zip(*sorted(enumerate(hm_scores), key=itemgetter(1), reverse=True))
        retro_scores = [pred.score for pred in candidate_text_element_predictions]
        r_indices, r_sorted = zip(*sorted(enumerate(retro_scores), key=itemgetter(1), reverse=True))
        n_elements = len(hm_scores)

        scored_elements = set()
        combined_scores = [0] * n_elements
        highest_score = n_elements
        hm_ind = 0
        r_ind = 0
        while len(scored_elements) < n_elements:
            while hm_indices[hm_ind] in scored_elements:
                hm_ind += 1
            combined_scores[hm_indices[hm_ind]] = highest_score
            scored_elements.add(hm_indices[hm_ind])
            highest_score -= 1
            hm_ind += 1
            if len(scored_elements) < n_elements:
                while r_indices[r_ind] in scored_elements:
                    r_ind += 1
                combined_scores[r_indices[r_ind]] = highest_score
                scored_elements.add(r_indices[r_ind])
                highest_score -= 1
                r_ind += 1
        return [x/n_elements for x in combined_scores]
